""" Unit Tests for Py-ART's io/nexrad_level3.py module. """

import numpy as np
from numpy.ma.core import MaskedArray

import pyart


def test_nexrad_level3_msg19():
    radar = pyart.io.read_nexrad_level3(pyart.testing.NEXRAD_LEVEL3_MSG19)

    assert radar.time['units'] == 'seconds since 2015-01-02T02:05:28Z'
    assert radar.time['data'].shape == (360, )
    assert round(radar.time['data'][0]) == 0.

    assert radar.range['data'].shape == (230, )
    assert round(radar.range['data'][100]) == 99900

    assert radar.scan_type == 'ppi'

    assert radar.latitude['data'].shape == (1, )
    assert round(radar.latitude['data'][0]) == 33.0

    assert radar.longitude['data'].shape == (1, )
    assert round(radar.longitude['data'][0]) == -87.0

    assert radar.altitude['data'].shape == (1, )
    assert round(radar.altitude['data'][0]) == 759.

    assert radar.altitude_agl is None

    assert radar.sweep_number['data'].shape == (1, )
    assert radar.sweep_number['data'][0] == 0

    assert radar.sweep_mode['data'].shape == (1, )
    assert radar.sweep_mode['data'].dtype.char == 'S'
    assert np.all(radar.sweep_mode['data'] == [b'azimuth_surveillance'])

    assert radar.fixed_angle['data'].shape == (1, )
    assert round(radar.fixed_angle['data'][0], 2) == 0.50

    assert radar.sweep_start_ray_index['data'].shape == (1, )
    assert round(radar.sweep_start_ray_index['data'][0]) == 0.0

    assert radar.sweep_end_ray_index['data'].shape == (1, )
    assert round(radar.sweep_end_ray_index['data'][0]) == 359.0

    assert radar.target_scan_rate is None

    assert round(radar.azimuth['data'][0]) == 320.0
    assert round(radar.azimuth['data'][10]) == 330.0

    assert radar.elevation['data'].shape == (360, )
    assert round(radar.elevation['data'][0], 2) == 0.50

    assert radar.scan_rate is None
    assert radar.antenna_transition is None
    assert radar.instrument_parameters is None
    assert radar.radar_calibration is None

    assert radar.ngates == 230
    assert radar.nrays == 360
    assert radar.nsweeps == 1

    assert 'reflectivity' in radar.fields.keys()
    assert radar.fields['reflectivity']['data'].shape == (360, 230)
    assert type(radar.fields['reflectivity']['data']) is MaskedArray
    assert round(radar.fields['reflectivity']['data'][10, 10]) == 25.


def test_nexrad_level3_msg161():
    radar = pyart.io.read_nexrad_level3(pyart.testing.NEXRAD_LEVEL3_MSG163)

    assert radar.time['units'] == 'seconds since 2015-01-02T02:05:28Z'
    assert radar.time['data'].shape == (360, )
    assert round(radar.time['data'][0]) == 0.

    assert radar.range['data'].shape == (1200, )
    assert round(radar.range['data'][100]) == 24975.

    assert radar.scan_type == 'ppi'

    assert radar.latitude['data'].shape == (1, )
    assert round(radar.latitude['data'][0]) == 33.0

    assert radar.longitude['data'].shape == (1, )
    assert round(radar.longitude['data'][0]) == -87.0

    assert radar.altitude['data'].shape == (1, )
    assert round(radar.altitude['data'][0]) == 759.

    assert radar.altitude_agl is None

    assert radar.sweep_number['data'].shape == (1, )
    assert radar.sweep_number['data'][0] == 0

    assert radar.sweep_mode['data'].shape == (1, )
    assert np.all(radar.sweep_mode['data'] == [b'azimuth_surveillance'])

    assert radar.fixed_angle['data'].shape == (1, )
    assert round(radar.fixed_angle['data'][0], 2) == 0.50

    assert radar.sweep_start_ray_index['data'].shape == (1, )
    assert round(radar.sweep_start_ray_index['data'][0]) == 0.0

    assert radar.sweep_end_ray_index['data'].shape == (1, )
    assert round(radar.sweep_end_ray_index['data'][0]) == 359.0

    assert radar.target_scan_rate is None

    assert round(radar.azimuth['data'][0]) == 329.0
    assert round(radar.azimuth['data'][10]) == 339.0

    assert radar.elevation['data'].shape == (360, )
    assert round(radar.elevation['data'][0], 2) == 0.50

    assert radar.scan_rate is None
    assert radar.antenna_transition is None
    assert radar.instrument_parameters is None
    assert radar.radar_calibration is None

    assert radar.ngates == 1200
    assert radar.nrays == 360
    assert radar.nsweeps == 1

    field_name = 'specific_differential_phase'
    assert field_name in radar.fields.keys()
    assert radar.fields[field_name]['data'].shape == (360, 1200)
    assert type(radar.fields[field_name]['data']) is MaskedArray
    assert round(radar.fields[field_name]['data'][103, 170]) == 2.


def test_nexrad_level3_msg161_fileobj():
    fh = open(pyart.testing.NEXRAD_LEVEL3_MSG163, 'rb')
    radar = pyart.io.read_nexrad_level3(fh)
    fh.close()
    field_name = 'specific_differential_phase'
    assert field_name in radar.fields.keys()
    assert radar.fields[field_name]['data'].shape == (360, 1200)
    assert type(radar.fields[field_name]['data']) is MaskedArray
    assert round(radar.fields[field_name]['data'][103, 170]) == 2.



""" Unit Tests for Py-ART's retrieve/simple_moment_calculation.py module. """

import numpy as np
import pyart


def test_calculate_snr_from_reflectivity():
    test_radar = pyart.testing.make_empty_ppi_radar(100, 360, 5)
    test_radar.range['data'] = test_radar.range['data'] * 100.0
    range_grid = np.meshgrid(
        test_radar.range['data'],
        np.ma.ones(test_radar.time['data'].shape))[0] + 1.0
    foo_field = {'data': np.zeros(
        [360 * 5, 100]) + 20.0 * np.log10(range_grid / 1000.)}
    test_radar.add_field('reflectivity', foo_field)
    snr = pyart.retrieve.calculate_snr_from_reflectivity(test_radar, toa=500)
    assert snr['data'].mean() < 1e-6


def test_calculate_velocity_texture():
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 1)

    # a zero field
    fdata = np.tile(np.arange(10.), 36).reshape(36, 10)
    fdata3 = np.zeros(fdata.shape)
    radar.add_field('zero_field', {'data': fdata3})

    vel_field = 'zero_field'
    texture_field = pyart.retrieve.calculate_velocity_texture(
        radar, vel_field, wind_size=2, nyq=10)
    assert np.all(texture_field['data'] == 0)
    texture_field = pyart.retrieve.calculate_velocity_texture(
        radar, vel_field, wind_size=3, nyq=10)
    assert np.all(texture_field['data'] == 0)
    texture_field = pyart.retrieve.calculate_velocity_texture(
        radar, vel_field, wind_size=4, nyq=10)
    assert np.all(texture_field['data'] == 0)



"""
====================================
Create a grid plot with data overlay
====================================

An example which creates a plot of a gridded NEXRAD radar on a map
with latitude and NCEP North American regional reanalysis (NARR) pressure
is plotted on top of the grid.

"""
print(__doc__)

# Author Jonathan J. Helmus, Cory Weber
# License: BSD 3 clause

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from netCDF4 import num2date, date2num, Dataset
import pyart


# read in the NEXRAD data, create the display
fname = '20110520100000_nexrad_grid.nc'
grid = pyart.io.read_grid(fname)
display = pyart.graph.GridMapDisplayBasemap(grid)

# create the figure
font = {'size': 10}
matplotlib.rc('font', **font)
fig = plt.figure(figsize=[10, 8])

# Add Basic Title
title = 'Basic Plot with Overlay Example Title'
#     Xleft%, ybot%
fig.text(0.5, 0.9, title, horizontalalignment='center', fontsize=24)

# panel sizes      xleft%, ybot%, xright% ,ytop%
map_panel_axes = [0.05, 0.15, 0.9, 0.7]
colorbar_panel_axes = [0.15, 0.09, 0.7, .010]

# parameters
level = 5
vmin = -8
vmax = 64
lat = 36.5
lon = -98.5

# panel 1, basemap, radar reflectivity and NARR overlay
ax1 = fig.add_axes(map_panel_axes)
display.plot_grid('REF', level=level, vmin=vmin, vmax=vmax, title_flag=False,
                  colorbar_flag=False)

# load overlay data
url = 'narr-a_221_20110520_0000_000.nc'
data = Dataset(url)

# extract data at correct time
grid_date = num2date(grid.time['data'], grid.time['units'])[0]
data_time = data.variables['time']
t_idx = abs(data_time[:] - date2num(grid_date, data_time.units)).argmin()
prmsl = 0.01 * data.variables['prmsl'][t_idx]

# plot the reanalysis on the basemap
lons, lats = np.meshgrid(data.variables['lon'], data.variables['lat'][:])

x, y = display.basemap(lons, lats)
clevs = np.arange(900, 1100., 1.)

display.basemap.contour(x, y, prmsl, clevs, colors='k', linewidths=1.)

# colorbar
cbax = fig.add_axes(colorbar_panel_axes)
display.plot_colorbar(cax=cbax)

# enable below to add cross hairs
# display.plot_crosshairs(lon=lon, lat=lat)

plt.show()



""" Cython Declarations for NASA TRMM RSL rsl.h file """

cdef extern from "rsl.h":
   
    cdef char[6] _RSL_VERSION_STR "RSL_VERSION_STR"

    # data structures
    ctypedef unsigned short Range
    
    ctypedef struct Ray_header:
        int   month 
        int   day  
        int   year  
        int   hour  
        int   minute
        float sec 
        float unam_rng 
        float azimuth 
        int   ray_num   
        float elev    
        int   elev_num  
        int   range_bin1
        int   gate_size 
        float  vel_res  
        float sweep_rate
        int prf         
        int prf2        
        float azim_rate 
        float fix_angle 
        float pitch
        float roll      
        float heading   
        float pitch_rate
        float roll_rate 
        float heading_rate
        float lat
        float lon       
        int   alt      
        float rvc     
        float vel_east
        float vel_north
        float vel_up    
        int   pulse_count 
        float pulse_width 
        float beam_width 
        float frequency   
        float wavelength  
        float nyq_vel  
        float (*f)(Range x)     
        Range (*invf)(float x) 
        int   nbins

    ctypedef struct Ray:
        Ray_header h
        Range * range

    ctypedef struct Sweep_header:
        int sweep_num         
        float elev
        float azimuth 
        float beam_width 
        float vert_half_bw 
        float horz_half_bw
        int nrays
        float (*f)(Range x) 
        Range (*invf)(float x)

    ctypedef struct Sweep:           
        Sweep_header h   
        Ray **ray

    ctypedef struct Volume_header:
        char *type_str 
        int nsweeps
        float calibr_const   
        float (*f)(Range x)   
        Range (*invf)(float x) 

    ctypedef struct Volume:
        Volume_header h
        Sweep **sweep

    ctypedef struct Radar_header:
        int month, day, year
        int hour, minute
        float sec
        char radar_type[50]
        int nvolumes
        int number     
        char name[8]     
        char radar_name[8]
        char project[24] 
        char city[15]  
        char state[3]    
        char country[15]
        int latd
        int latm 
        int lats  
        int lond  
        int lonm  
        int lons   
        int height 
        int spulse 
        int lpulse
        int scan_mode 
        int vcp

    ctypedef struct Radar:
        Radar_header h
        Volume **v
    
    # function definitions
    Radar * RSL_anyformat_to_radar(char *infile)
    Radar * RSL_wsr88d_to_radar(char *infile, char *call_or_first_tape_file)
    
    Volume * RSL_copy_volume(Volume *v)
    
    Volume * RSL_new_volume(int max_sweeps)
    Sweep * RSL_new_sweep(int max_rays) 
    Ray * RSL_new_ray(int max_bins)
   
    void RSL_free_ray(Ray *r)
    void RSL_free_sweep(Sweep *s)
    void RSL_free_volume(Volume *v)
    void RSL_free_radar(Radar *r)
    
    void RSL_print_version()
    
    # conversion functions
    float DZ_F(Range x)
    float VR_F(Range x)
    float SW_F(Range x)
    float CZ_F(Range x)
    float ZT_F(Range x)
    float DR_F(Range x)
    float LR_F(Range x)
    float ZD_F(Range x)
    float DM_F(Range x)
    float RH_F(Range x)
    float PH_F(Range x)
    float XZ_F(Range x)
    float CD_F(Range x)
    float MZ_F(Range x)
    float MD_F(Range x)
    float ZE_F(Range x)
    float VE_F(Range x)
    float KD_F(Range x)
    float TI_F(Range x)
    float DX_F(Range x)
    float CH_F(Range x)
    float AH_F(Range x)
    float CV_F(Range x)
    float AV_F(Range x)
    float SQ_F(Range x)
    float VS_F(Range x)
    float VL_F(Range x)
    float VG_F(Range x)
    float VT_F(Range x)
    float NP_F(Range x)
    float HC_F(Range x)
    float VC_F(Range x)
    float SD_F(Range x)
    
    Range DZ_INVF(float x)
    Range VR_INVF(float x)
    Range SW_INVF(float x)
    Range CZ_INVF(float x)
    Range ZT_INVF(float x)
    Range DR_INVF(float x)
    Range LR_INVF(float x)
    Range ZD_INVF(float x)
    Range DM_INVF(float x)
    Range RH_INVF(float x)
    Range PH_INVF(float x)
    Range XZ_INVF(float x)
    Range CD_INVF(float x)
    Range MZ_INVF(float x)
    Range MD_INVF(float x)
    Range ZE_INVF(float x)
    Range VE_INVF(float x)
    Range KD_INVF(float x)
    Range TI_INVF(float x)
    Range DX_INVF(float x)
    Range CH_INVF(float x)
    Range AH_INVF(float x)
    Range CV_INVF(float x)
    Range AV_INVF(float x)
    Range SQ_INVF(float x)
    Range VS_INVF(float x)
    Range VL_INVF(float x)
    Range VG_INVF(float x)
    Range VT_INVF(float x)
    Range NP_INVF(float x)
    Range HC_INVF(float x)
    Range VC_INVF(float x)
    Range SD_INVF(float x)



""" Unit Tests for Py-ART's retrieve/advection.py module. """

from numpy.testing import assert_almost_equal
import pyart


def test_grid_displacement_pc():
    grid1 = pyart.testing.make_storm_grid()
    data = grid1.fields['reflectivity']['data'].copy()
    grid1.fields['reflectivity']['data'] = data[:, 5:-5, 3:-3].copy()

    grid2 = pyart.testing.make_storm_grid()
    grid2.fields['reflectivity']['data'] = data[:, 0:-10, 0:-6].copy()

    # test pixels
    displacement = pyart.retrieve.grid_displacement_pc(
        grid1, grid2, 'reflectivity', 0)
    assert displacement == (-5, -3)

    # test distance
    grid1.fields['reflectivity']['valid_min'] = 0
    grid2.fields['reflectivity']['valid_min'] = 0
    displacement = pyart.retrieve.grid_displacement_pc(
        grid1, grid2, 'reflectivity', 0, return_value='distance')
    dx = grid1.x['data'][1] - grid1.x['data'][0]
    dy = grid1.y['data'][1] - grid1.y['data'][0]
    assert_almost_equal(displacement[0], -5*dy, 0)
    assert_almost_equal(displacement[1], -3*dx, 0)

    # test velocity
    grid2.time['data'] += 1
    displacement = pyart.retrieve.grid_displacement_pc(
        grid1, grid2, 'reflectivity', 0, return_value='velocity')
    assert_almost_equal(displacement[0], -5*dy, 0)
    assert_almost_equal(displacement[1], -3*dx, 0)

    # test fallback
    displacement = pyart.retrieve.grid_displacement_pc(
        grid1, grid2, 'reflectivity', 0, return_value='foobar')
    assert displacement == (-5, -3)


def test_grid_shift():

    # create two guassian storms
    grid1 = pyart.testing.make_normal_storm(10.0, [0.0, 0.0])
    grid2 = pyart.testing.make_normal_storm(10.0, [5.0, 5.0])

    # trim one, trim and shift the other
    trimmed_grid2 = pyart.retrieve.grid_shift(grid2, [0.0, 0.0], trim_edges=10)
    shifted_grid1 = pyart.retrieve.grid_shift(grid1, [5.0, 5.0], trim_edges=10)

    # The difference should be nearly zero
    data1 = shifted_grid1.fields['reflectivity']['data'][0]
    data2 = trimmed_grid2.fields['reflectivity']['data'][0]
    assert (data1 - data2).mean() < 1.e-10



#!/opt/local/opkg/bin/python
"""
metranet library
================

functions to read METRANET files, require external shared library
srn_idl_py_lib.<ARCH>.so

.. autosummary::
    :toctree: generated/

    get_radar_site_info
    get_library_path
    get_library
    read_polar
    read_product
    read_file


.. autosummary::
    :toctree: generated/
    :template: dev_template.rst

    RadarData
    AzimuthHeader_stru
    SweepHeader_stru
    Selex_Angle

"""

from __future__ import print_function

import ctypes
import os
import sys
import platform
import string
import time
from warnings import warn
import traceback

import numpy as np

from .dn_to_float import float_mapping_m, float_mapping_p, nyquist_vel

# some values valid for all sites
NPM_MOM = 11
NPH_MOM = 12
NPL_MOM = 9

# fix for python3
if sys.version_info[0] == 3:
    def xrange(i):
        return range(i)


class RadarData:
    """
    A class for storing radar data.

    Attributes
    ----------
    type : str
        type of data
    data : numpy array or numpy masked array
        array containing the data
    scale : numpy array
        array containing the scale used to transform the data from digital
        to physical units
    header : dict
        dictionary containing metadata
    pol_header : dict
        dictionary containing metadata of the polar files
    moment : str
        moment name

    """
    type = "RadarData"

    def __init__(self, data=np.zeros(0),
                 scale=np.fromiter(xrange(256), dtype=np.uint), header=(),
                 pol_header=(), moment='ZH'):
        self.data = data
        self.scale = scale
        self.header = header
        self.pol_header = pol_header
        self.moment = moment


class AzimuthHeader_stru(ctypes.Structure):
    """
    A class containing the data from the header of the polar PM files

    Attributes
    ----------
    _fields_: dict
        A dictionary containing the metadata contained in the PM file

    C-Structure of METRANET POLAR data PM format

    struct moment_header_struct
    {
            unsigned int record_type;  data format (moment1) + moment mask
            unsigned int scan_id;
            unsigned int host_id;
            unsigned int start_angle;
            unsigned int end_angle;

            unsigned char ant_mode;
            unsigned char total_sweep;
            unsigned char current_sweep;  1-any number up to 99
            unsigned char end_of_sweep;  0=not end, 1=end sweep, 2=end volume

            short sequence;  ray sequence number in a sweep
            short total_record;  total ray number in sweep
            short pulses;
            short num_gates;

            int data_bytes;
            unsigned short data_flag;
            short data_time_residue;  data time residue in 0.01 sec
            unsigned int data_time;  data time in second
            short repeat_time;
            char compressed;  flag for compression of data
            char priority;  for file name use

            float   ny_quest;
            float   gate_width;
            float   w_ny_quest;  may be used for other variable
            float   start_range;
    };
    """
    _fields_ = [
        ("record_type", ctypes.c_uint),
        ("scan_id", ctypes.c_int),
        ("host_id", ctypes.c_int),
        ("start_angle", ctypes.c_int),
        ("end_angle", ctypes.c_int),

        ("ant_mode", ctypes.c_ubyte),
        ("total_sweep", ctypes.c_ubyte),
        ("current_sweep", ctypes.c_ubyte),
        ("end_of_sweep", ctypes.c_ubyte),

        ("sequence", ctypes.c_short),
        ("total_record", ctypes.c_short),
        ("pulses", ctypes.c_short),
        ("num_gates", ctypes.c_short),

        ("data_bytes", ctypes.c_int),
        ("data_flag", ctypes.c_ushort),
        ("data_time_residue", ctypes.c_short),
        ("data_time", ctypes.c_uint),
        ("repeat_time", ctypes.c_short),
        ("compressed", ctypes.c_ubyte),
        ("priority", ctypes.c_ubyte),

        ("ny_quest", ctypes.c_float),
        ("gate_width", ctypes.c_float),
        ("w_ny_quest", ctypes.c_float),
        ("start_range", ctypes.c_float),
    ]

class FileHeader_stru(ctypes.Structure):
    """
    A class containing the data from the header of the polar Mx files

    Attributes
    ----------
    _fields_: dict
        A dictionary containing the metadata contained in the Mx file

    C-Structure of METRANET POLAR data MS format

    C-code from metranet.git/share/include/prd_header.h
{
             int  row;                        /* polar: 360, rect: Y   */
             int  column;                     /* polar: gate, rect: X  */
             unsigned int data_time;
             float data_elevation;            /* elevation angle or the cappi height */
             float data_width;                /* gate width for polar        */
             float nyquist;                   /* number of nyquist speed */
             int   current_sweep;             /* polar or cappi level number */
             int   total_sweep;               /* polar or cappi level number */
             int   ant_mode;
             float w_nyquist;
             int repeat_time;
             int data_flags;                  /* for scan mode               */

             int data_mom_type;               /* data format type 1, 2, 3    */
             int data_mom_bytes;              /* data moment bytes  1, 4, 2  */
             float scale;
             float offset;
             float factorc;
             int scale_type;
             unsigned int data_end_time;
             float data_azimuth;
             int priority;
             int quality;
             float wave_length;
             float pulse_width;
             float radar_lat;
             float radar_lon;
             float radar_h;
             char moment_name[16];
             char moment_unit[16];
             char radar_name[16];
             char scan_name[16];
        };
    """
    _fields_ = [
        ("row", ctypes.c_int32),
        ("column", ctypes.c_int32),
        ("data_time", ctypes.c_uint32),
        ("data_elevation", ctypes.c_float),
        ("data_width", ctypes.c_float),
        ("nyquist", ctypes.c_float),
        ("current_sweep", ctypes.c_int32),
        ("total_sweep", ctypes.c_int32),
        ("ant_mode", ctypes.c_int32),
        ("w_nyquist", ctypes.c_float),
        ("repeat_time", ctypes.c_int32),
        ("data_flags", ctypes.c_int32),
        ("data_mom_type", ctypes.c_int32),
        ("data_mom_bytes", ctypes.c_int32),
        ("scale", ctypes.c_float),
        ("offset", ctypes.c_float),
        ("factorc", ctypes.c_float),
        ("scale_type", ctypes.c_int32),
        ("data_end_time", ctypes.c_uint32),
        ("data_azimuth", ctypes.c_float),
        ("priority", ctypes.c_int32),
        ("quality", ctypes.c_int32),
        ("wave_length", ctypes.c_float),
        ("pulse_width", ctypes.c_float),
        ("radar_lat", ctypes.c_float),
        ("radar_lon", ctypes.c_float),
        ("radar_h", ctypes.c_float),
        ("moment_name", ctypes.c_char*16),
        ("moment_unit", ctypes.c_char*16),
        ("radar_name", ctypes.c_char*16),
        ("scan_name", ctypes.c_char*16),

    ]


class SweepHeader_stru(ctypes.Structure):
    """
    A class containing the data from the header of the polar MS files

    Attributes
    ----------
    _fields_: dict
        A dictionary containing the metadata contained in the MS file

    C-Structure of METRANET POLAR data MS format

    C-code from METRANET2/share/include/sweep_file.h

    struct sweep_header
        {
            int8_t FileId[4];           //4:4
            uint8_t Version;            //1:5
            uint8_t Spare1[3];          //3:8
            uint32_t Length;            //4:12
            int8_t RadarName[16];       //16:28
            int8_t ScanName[16];        //16:44
            float RadarLat;             //4:48
            float RadarLon;             //4:52
            float RadarHeight;          //4:56
            uint8_t SequenceSweep;      //1:57
            uint8_t CurrentSweep;       //1:58
            uint8_t TotalSweep;         //1:59
            uint8_t AntMode;            //1:60
            uint8_t Priority;           //1:61
            uint8_t Quality;            //1:62
            uint8_t Spare2[2];          //2:64
            uint16_t RepeatTime;        //2:66
            uint16_t NumMoments;        //2:68
            float GateWidth;            //4:72
            float WaveLength;           //4:76
            float PulseWidth;           //4:80
            float StartRange;           //4:84
            uint32_t MetaDataSize;      //4:88
        };
    """
    _fields_ = [
        ("FileId", ctypes.c_int32),
        ("Version", ctypes.c_uint8),
        ("Spare1", ctypes.c_uint8*3),
        ("Length", ctypes.c_uint32),
        ("RadarName", ctypes.c_int8*16),
        ("ScanName", ctypes.c_int8*16),
        ("RadarLat", ctypes.c_float),
        ("RadarLon", ctypes.c_float),
        ("RadarHeight", ctypes.c_float),
        ("SequenceSweep", ctypes.c_uint8),
        ("CurrentSweep", ctypes.c_uint8),
        ("TotalSweep", ctypes.c_uint8),
        ("AntMode", ctypes.c_uint8),
        ("Priority", ctypes.c_uint8),
        ("Quality", ctypes.c_uint8),
        ("Spare2", ctypes.c_uint8*2),
        ("RepeatTime", ctypes.c_uint16),
        ("NumMoments", ctypes.c_uint16),
        ("GateWidth", ctypes.c_float),
        ("WaveLength", ctypes.c_float),
        ("PulseWidth", ctypes.c_float),
        ("StartRange", ctypes.c_float),
        ("MetaDataSize", ctypes.c_uint32),
    ]


class Selex_Angle:
    """
    Class used to convert from digital number to angle

    Attributes
    ----------
    az : float
        azimuth angle value (degrees or radiants)
    el : float
        elevation angle value (degrees or radiants)

    """
    def __init__(self, angle=0, radiant=False):
        if radiant:
            reform = 2 * 3.1415926
        else:
            reform = 360.
        self.az = (angle & 0xFFFF)/65535.*reform
        self.el = (angle >> 16)/65535.*reform


def get_radar_site_info(verbose=False):
    """
        return dictionary with radar'info

        Returns
        -------
        radar_def : dict
            dictionary containing radar site information
    """
    radar_def_load = False
    try:
        import yaml
        yaml_module = True
    except ImportError:
        yaml_module = False

    if yaml_module:
        metranet_yaml_file = "metranet.yaml"
        path_yaml_file = [
            "/store/msrad/python/library/radar/io/", "/opt/ccs4/python",
            "/proj/lom/python/library/radar/io"]
        for p in path_yaml_file:
            full_file = p + "/" + metranet_yaml_file
            if os.path.isfile(full_file):
                print("look for Radar_Site_info from %s" % full_file)
                try:
                    with open(full_file) as f:
                        radar_def = yaml.load(f)
                        radar_def_load = True
                        if verbose:
                            print("Read Radar_Site_info from %s" % full_file)
                except OSError as ee:
                    warn(str(ee))
                    traceback.print_exc()

    if not radar_def_load:
        # HardCoded definition
        print("**** HARD-CODED Radar site Value, may be not updated!!")
        # some values valid for all sites
        c_speed = 299792458  # m/s

        radar_default = {}

        radar_default['dwhname'] = 'undef'
        radar_default['RadarCHY'] = float('nan')
        radar_default['RadarCHX'] = float('nan')
        radar_default['FileId'] = 1179665477
        radar_default['Version'] = 0
        radar_default['RadarName'] = "undef"
        radar_default['ScanName'] = "undef"
        radar_default['RadarLat'] = float('nan')
        radar_default['RadarLon'] = float('nan')
        radar_default['RadarHeight'] = float('nan')
        radar_default['Frequency'] = float('nan')
        radar_default['WaveLength'] = float('nan')
        radar_default['PulseWidth'] = 0.5
        radar_default['SweepsOrder'] = (
            9, 7, 5, 3, 1, 19, 17, 15, 13, 11, 10, 8, 6, 4, 2, 20, 18, 16, 14,
            12)
        radar_default['NumMomentsPM'] = NPM_MOM
        radar_default['NumMomentsPH'] = NPH_MOM
        radar_default['NumMomentsPL'] = NPL_MOM

        radar_def = {}

        rname = 'A'
        radar_def[rname] = radar_default.copy()
        radar_def[rname]['dwhname'] = "ALB"
        radar_def[rname]['RadarName'] = "Albis"
        radar_def[rname]['RadarCHY'] = 681201
        radar_def[rname]['RadarCHX'] = 237604
        radar_def[rname]['RadarLat'] = 47.284333
        radar_def[rname]['RadarLon'] = 8.512000
        radar_def[rname]['RadarHeight'] = 938.0
        radar_def[rname]['ScanName'] = "1095516672"
        radar_def[rname]['Frequency'] = 5450e6
        radar_def[rname]['WaveLength'] = (
            c_speed/radar_def[rname]['Frequency']*1e2)

        rname = 'D'
        radar_def[rname] = radar_default.copy()
        radar_def[rname]['dwhname'] = "DOL"
        radar_def[rname]['RadarName'] = "Dole"
        radar_def[rname]['RadarCHY'] = 497057
        radar_def[rname]['RadarCHX'] = 142408
        radar_def[rname]['RadarLat'] = 46.425113
        radar_def[rname]['RadarLon'] = 6.099415
        radar_def[rname]['RadarHeight'] = 1682.0
        radar_def[rname]['ScanName'] = "1146047488"
        radar_def[rname]['Frequency'] = 5430e6
        radar_def[rname]['WaveLength'] = (
            c_speed/radar_def[rname]['Frequency']*1e2)

        rname = 'L'
        radar_def[rname] = radar_default.copy()
        radar_def[rname]['dwhname'] = "MLE"
        radar_def[rname]['RadarName'] = "Lema"
        radar_def[rname]['RadarCHY'] = 707957
        radar_def[rname]['RadarCHX'] = 99762
        radar_def[rname]['RadarLat'] = 46.040761
        radar_def[rname]['RadarLon'] = 8.833217
        radar_def[rname]['RadarHeight'] = 1626.0
        radar_def[rname]['ScanName'] = "1279610112"
        radar_def[rname]['Frequency'] = 5455e6
        radar_def[rname]['WaveLength'] = (
            c_speed/radar_def[rname]['Frequency']*1e2)

        rname = 'P'
        radar_def[rname] = radar_default.copy()
        radar_def[rname]['dwhname'] = "PPM"
        radar_def[rname]['RadarName'] = "PlaineMorte"
        radar_def[rname]['RadarCHY'] = 603687
        radar_def[rname]['RadarCHX'] = 135476
        radar_def[rname]['RadarLat'] = 46.370646
        radar_def[rname]['RadarLon'] = 7.486552
        radar_def[rname]['RadarHeight'] = 2937.0
        radar_def[rname]['ScanName'] = "0"
        radar_def[rname]['Frequency'] = 5468e6
        radar_def[rname]['WaveLength'] = (
            c_speed/radar_def[rname]['Frequency']*1e2)

        rname = 'W'
        radar_def[rname] = radar_default.copy()
        radar_def[rname]['dwhname'] = "WEI"
        radar_def[rname]['RadarName'] = "WeissFluhGipfel"
        radar_def[rname]['RadarCHY'] = 779700
        radar_def[rname]['RadarCHX'] = 189790
        radar_def[rname]['RadarLat'] = 47.284333
        radar_def[rname]['RadarLon'] = 9.794458
        radar_def[rname]['RadarHeight'] = 2850.0
        radar_def[rname]['ScanName'] = "0"
        radar_def[rname]['Frequency'] = 5433e6
        radar_def[rname]['WaveLength'] = (
            c_speed/radar_def[rname]['Frequency']*1e2)

    return radar_def


def get_library_path():
    """
        look for valid library path

        Returns
        -------
        library_metranet_path : str
            METRANET library path
    """
    libray_paths = [
        os.environ.get('METRANETLIB_PATH'),
        "/store/msrad/idl/lib/radlib4", "/opt/ccs4/lib",
        "/opt/local/opkg/share/ccs4/lib",
        "/proj/lom/idl/lib/radlib4"]

    library_metranet_path = ''
    for p in libray_paths:
        if p is not None:
            if os.path.isdir(p):
                library_metranet_path = p
                break

    if not library_metranet_path:
        sys.exit(" ENV Variable library_metranet_path NOT defined")

    return library_metranet_path


def get_library(verbose=False, momentpm=False, momentms=True):
    """
        return the link to C-shared library

        Parameters
        ----------
        verbose : Boolean
            If true print out extra information
        momentspm : Boolean
            If true returns the link to the MS library
        momentsms : Boolean
            If true returns the link to the MS library
        Returns
        -------
        metranet_lib : link
            loaded METRANET C-library

    """
    # default path
    try:
        library_metranet_path = get_library_path()
    except SystemExit:
        sys.exit('ERROR: METRANET library path not found')

    # library system
    library_metranet_linux = 'srn_idl_py_lib.x86_64'
    library_metranet_sparc32 = 'srn_idl_py_lib.sparc32.so'
    # library_metranet_sparc64 = 'srn_idl_py_lib.sparc64.so'

    library_metranet = 'x'
    if platform.system() == 'SunOS':
        if platform.architecture()[0] == '64bit':
            library_metranet = library_metranet_sparc32
        else:
            library_metranet = library_metranet_sparc32

    if platform.system() == 'Linux':
        if momentms:
            library_metranet = library_metranet_linux + '.MS.so'
        elif momentpm:
            library_metranet = library_metranet_linux + '.PM.so'
        else:
            library_metranet = library_metranet_linux + '.so'

    if verbose:
        print("library %s/%s:" % (library_metranet_path, library_metranet))

    if library_metranet == 'x':
        sys.exit("ERROR: Platform not found")

    metranet_lib = ctypes.cdll.LoadLibrary(
        library_metranet_path + '/' + library_metranet)

    return metranet_lib


def read_polar(radar_file, moment="ZH", physic_value=False,
               masked_array=False, verbose=False):
    """
    Reads a METRANET polar data file

    Parameters
    ----------
    radar_file : str
        file name
    moment : str
        moment name
    physic_value : boolean
        If true returns the physical value. Otherwise the digital value
    masked_array : boolean
        If true returns a numpy masked array with NaN values masked. Otherwise
        returns a regular masked array with NaN values
    verbose : boolean
        If true prints out extra information

    Returns
    -------
    ret_data : RadarData object
        An object containing the information read from the file. None if
        the file has not been properly read

    """
    c_speed = 299792458  # m/s

    ret_data = RadarData(moment=moment)
    prd_header = {}
    prd_data_level = ret_data.scale

    # uppercase moment
    moment = moment.upper()

    # as big as possible
    max_bins = 3000
    max_azimuths = 500

    if verbose:
        print("Read POLAR file %s " % radar_file)

        # read BINARY data
    prdt_size = max_bins * max_azimuths
    if moment == 'PHI':
        prdt_size *= 2
        prd_data = np.zeros(prdt_size, np.ushort)
    else:
        prd_data = np.zeros(prdt_size, np.ubyte)

    momentpm = False
    momentms = False
    bfile = os.path.basename(radar_file)
    if (bfile.startswith('MS') or bfile.startswith('MH') or
            bfile.startswith('ML')):
        momentms = True
    elif (bfile.startswith('PM') or bfile.startswith('PH') or
            bfile.startswith('PL')):
        momentpm = True

    t_pol_header = (AzimuthHeader_stru * max_azimuths)()
    t_rad_header = (SweepHeader_stru * 1)()
    t_all_header = FileHeader_stru()
    
    metranet_lib = get_library(momentms=momentms, momentpm=momentpm,
                               verbose=verbose)

    ret = metranet_lib.py_decoder_p2ext(
        ctypes.c_char_p(radar_file.encode('utf-8')),
        np.ctypeslib.as_ctypes(prd_data), ctypes.c_int(prdt_size),
        ctypes.c_char_p(moment.encode('utf-8')), ctypes.byref(t_pol_header),
        ctypes.byref(t_rad_header), ctypes.byref(t_all_header), 
        ctypes.c_int(verbose))

    if ret <= max_azimuths:
        return None

    if moment == 'PHI':
        ret *= 0.5

    # reshape matrix data
    bins = t_pol_header[0].num_gates
    nr_az = int(ret / bins)
    if bins < 1:
        # if num_gates is less than 1 (exception)
        bins = ret/360
    if nr_az > 360:
        nr_az = 360

    if nr_az < 360:
        print("WARNING incomplete sweep")

    if verbose:
        print("ret=%d" % ret)
        print("bins=%d" % bins)
        print("nr_az=%d" % nr_az)
        print("len(prd_data)=%d" % len(prd_data))
    prd_data = prd_data[0: nr_az * bins]
    prd_data = np.reshape(prd_data, (nr_az, bins))

    # reorder pol_header
    pol_header = (AzimuthHeader_stru * nr_az)()
    for i in range(0, nr_az):
        angle_start = Selex_Angle(t_pol_header[i].start_angle)
        pol_header[int(angle_start.az)] = t_pol_header[i]
        
    if momentms:
        # Select scale
        momhead = {
        'scale_type': t_all_header.scale_type,
        'a' : t_all_header.scale,
        'b' : t_all_header.offset,
        'c' : t_all_header.factorc}
    
        prd_data_level = float_mapping_m(moment, momhead, pol_header[0].data_time,
                                       pol_header[0].scan_id,
                                       pol_header[0].ny_quest)
    else:
        prd_data_level = float_mapping_p(
                        moment, pol_header[0].data_time,
                        pol_header[0].scan_id, pol_header[0].ny_quest)
         
    if verbose:
        print("prd_data shape ", prd_data.shape)
        print("min/max prd_data: ", prd_data.min(), prd_data.max())
        set(string.printable)

        print("prd_data scan_id ", pol_header[0].scan_id)
        print("prd_data host_id ", pol_header[0].host_id)
        print("data level ", prd_data_level[0:10])
        for i in range(0, nr_az, 10):
            angle_start = Selex_Angle(pol_header[i].start_angle)
            angle_end = Selex_Angle(pol_header[i].end_angle)
            x = pol_header[i].num_gates / 2
            print(("pol_header[%3d].num_gates: %d, time %d.%03d start_az/el:"
                   " %6.1f/%4.1f  end_az/el: %6.1f/%4.1f real[%d]=%6.2f"
                   " ( raw=%5d)") %
                  (i, pol_header[i].num_gates, pol_header[i].data_time,
                   pol_header[i].data_time_residue, angle_start.az,
                   angle_start.el, angle_end.az, angle_end.el, x,
                   prd_data_level[prd_data[i, x]], prd_data[i, x]))

    if physic_value:
        ret_data.data = prd_data_level[prd_data]
        if masked_array:
            ret_data.data = np.ma.array(
                ret_data.data, mask=np.isnan(ret_data.data))
            if bfile[1] == 'L':
                if moment in ('ZH', 'ZV', 'ZHC', 'ZVC'):
                    prd_data_level[1] = np.nan
                    ret_data.data = np.ma.masked_where(
                        prd_data == 1, ret_data.data)
    else:
        ret_data.data = prd_data
        if masked_array:
            ret_data.data = np.ma.array(
                ret_data.data, mask=prd_data == 0)
            if bfile[1] == 'L':
                if moment in ('ZH', 'ZV', 'ZHC', 'ZVC'):
                    prd_data_level[1] = np.nan
                    ret_data.data = np.ma.masked_where(
                        prd_data == 1, ret_data.data)

    # header
    prd_header['pid'] = bfile[0:3]
    prd_header['radar'] = bfile[2]
    prd_header['moment'] = moment
    prd_header['column'] = bins
    prd_header['row'] = nr_az
    # prd_header['elevation'] = "%03d" % pol_header[0].current_sweep
    prd_header['volume_time'] = int(
        pol_header[0].data_time - pol_header[0].data_time % 300 + 300)
    prd_header['time'] = time.strftime(
        '%y%j%H%M', time.gmtime(int(prd_header['volume_time'])))
    prd_header["quality"] = bfile[12]

    # if exists extended header, fill header
    if t_rad_header[0].FileId > 0 and t_rad_header[0].GateWidth > 0.:
        prd_header["FileId"] = t_rad_header[0].FileId
        prd_header["Version"] = t_rad_header[0].Version
        # prd_header["Spare1"] = t_rad_header[0].Spare1
        # prd_header["Length"] = t_rad_header[0].Length
        prd_header["RadarName"] = ctypes.string_at(
            t_rad_header[0].RadarName)
        prd_header["ScanName"] = ctypes.string_at(t_rad_header[0].ScanName)
        prd_header["RadarLat"] = t_rad_header[0].RadarLat
        prd_header["RadarLon"] = t_rad_header[0].RadarLon
        prd_header["RadarHeight"] = t_rad_header[0].RadarHeight
        prd_header["SequenceSweep"] = t_rad_header[0].SequenceSweep
        prd_header["CurrentSweep"] = t_rad_header[0].CurrentSweep
        prd_header["TotalSweep"] = t_rad_header[0].TotalSweep
        prd_header["AntMode"] = t_rad_header[0].AntMode
        prd_header["Priority"] = t_rad_header[0].Priority
        # prd_header["Quality"] = t_rad_header[0].Quality
        prd_header["quality"] = t_rad_header[0].Quality
        # prd_header["Spare2"] = t_rad_header[0].Spare2
        prd_header["RepeatTime"] = t_rad_header[0].RepeatTime
        prd_header["NumMoments"] = t_rad_header[0].NumMoments
        prd_header["GateWidth"] = t_rad_header[0].GateWidth
        prd_header["WaveLength"] = t_rad_header[0].WaveLength
        prd_header["Frequency"] = c_speed/(prd_header["WaveLength"]*1e-2)
        prd_header["PulseWidth"] = t_rad_header[0].PulseWidth
        prd_header["StartRange"] = t_rad_header[0].StartRange
        prd_header["MetaDataSize"] = t_rad_header[0].MetaDataSize
        # update with radar sweep radar
        prd_header['radar'] = prd_header["RadarName"][0]
    else:
        # specific radar metadata not available (old PM/PH/PL format)
        # from polar header
        prd_header["CurrentSweep"] = pol_header[0].current_sweep
        prd_header["TotalSweep"] = pol_header[0].total_sweep
        prd_header["AntMode"] = pol_header[0].ant_mode
        prd_header["RepeatTime"] = pol_header[0].repeat_time
        prd_header["GateWidth"] = pol_header[0].gate_width
        prd_header["StartRange"] = pol_header[0].start_range
        prd_header["Priority"] = pol_header[0].priority

        # from hard-coded table
        rname = prd_header['radar']
        radar_def = get_radar_site_info()
        if rname in radar_def.keys():
            prd_header["FileId"] = radar_def[rname]["FileId"]
            prd_header["Version"] = radar_def[rname]["Version"]
            prd_header["RadarName"] = radar_def[rname]["RadarName"]
            prd_header["ScanName"] = radar_def[rname]["ScanName"]
            prd_header["RadarLat"] = radar_def[rname]["RadarLat"]
            prd_header["RadarLon"] = radar_def[rname]["RadarLon"]
            prd_header["RadarHeight"] = radar_def[rname]["RadarHeight"]
            prd_header["Frequency"] = radar_def[rname]["Frequency"]
            prd_header["WaveLength"] = radar_def[rname]["WaveLength"]
            prd_header["PulseWidth"] = radar_def[rname]["PulseWidth"]
            prd_header["SequenceSweep"] = (
                radar_def[rname]['SweepsOrder'].index(
                    pol_header[0].current_sweep))

            if bfile[1] == 'M':
                prd_header["NumMoments"] = radar_def[rname]['NumMomentsPM']
            if bfile[1] == 'H':
                prd_header["NumMoments"] = radar_def[rname]['NumMomentsPH']
            if bfile[1] == 'L':
                prd_header["NumMoments"] = radar_def[rname]['NumMomentsPL']

    ret_data.header = prd_header
    ret_data.scale = prd_data_level
    ret_data.pol_header = pol_header

    # change parameters in header
    for i in range(nr_az):
        ret_data.pol_header[i].total_record = nr_az

    return ret_data


def read_product(radar_file, physic_value=False, masked_array=False,
                 verbose=False):
    """
    Reads a METRANET cartesian data file

    Parameters
    ----------
    radar_file : str
        file name
    physic_value : boolean
        If true returns the physical value. Otherwise the digital value
    masked_array : boolean
        If true returns a numpy masked array with NaN values masked. Otherwise
        returns a regular masked array with NaN values
    verbose : boolean
        If true prints out extra information

    Returns
    -------
    ret_data : RadarData object
        An object containing the information read from the file. None if
        the file has not been properly read

    """
    ret_data = RadarData()
    prd_header = {'row': 0, 'column': 0}

    # read ASCII data
    if verbose:
        print("physic_value: ", physic_value)
        print("File %s: read ASCII" % radar_file)

    try:
        with open(radar_file, 'rb') as data_file:
            lines = data_file.readlines()
    except OSError as ee:
        warn(str(ee))
        print("Unable to read file '%s'" % radar_file)
        return None

    for t_line in lines:
        line = t_line.decode("utf-8").strip('\n')
        if line.find('end_header') == -1:
            data = line.split('=')
            prd_header[data[0]] = data[1]
        else:
            break

    # read BINARY data
    prdt_size = int(prd_header['column']) * int(prd_header['row'])
    if prdt_size < 1:
        print("Error, no size found row=%3d column=%3d" %
              (prd_header['row'], prd_header['column']))
        return None

    if verbose:
        print("File %s: read BINARY data: expected %s bytes, " %
              (radar_file, prdt_size), end='')

    prd_data = np.zeros(
        [int(prd_header['row']), int(prd_header['column'])], np.ubyte)
    prd_data_level = np.zeros(256, np.float32)
    metranet_lib = get_library(verbose=verbose)

    ret = metranet_lib.py_decoder(
        ctypes.c_char_p(radar_file.encode('utf-8')),
        np.ctypeslib.as_ctypes(prd_data), ctypes.c_int(prdt_size),
        np.ctypeslib.as_ctypes(prd_data_level), ctypes.c_int(verbose))

    # convert 0 at end of array with NAN
    conv_zero2nan = True

    i = len(prd_data_level)
    if prd_data_level.max() == prd_data_level.min():
        prd_data_level = np.fromiter(xrange(256), dtype=np.uint32)
    else:
        while conv_zero2nan:
            i -= 1
            if i < 0:
                conv_zero2nan = False
            elif prd_data_level[i] == 0.0:
                prd_data_level[i] = np.nan
            else:
                conv_zero2nan = False

    if verbose:
        print("Found %d bytes" % ret)
        print("prd_data_level[10] = %f" % prd_data_level[10])
        print("min/max prd_data: %d/%d" % (prd_data.min(), prd_data.max()))
        print("first 100 bytes", prd_data[0:100, 0])
        print("data level ", prd_data_level[0:10])

    # ret_data = RadarData(
    #    data=prd_data, header=prd_header, scale=prd_data_level)
    if physic_value:
        ret_data.data = prd_data_level[prd_data]
        if masked_array:
            ret_data.data = np.ma.array(
                ret_data.data, mask=np.isnan(ret_data.data))
            ret_data.data = np.ma.masked_where(prd_data == 0, ret_data.data)
    else:
        ret_data.data = prd_data
        if masked_array:
            ret_data.data = np.ma.array(
                ret_data.data, mask=prd_data == 0)
    ret_data.header = prd_header
    ret_data.scale = prd_data_level

    return ret_data


def read_file(file, moment="ZH", physic_value=False, masked_array=False,
              verbose=False):
    """
    Reads a METRANET data file

    Parameters
    ----------
    file : str
        file name
    moment : str
        moment name
    physic_value : boolean
        If true returns the physical value. Otherwise the digital value
    masked_array : boolean
        If true returns a numpy masked array with NaN values masked. Otherwise
        returns a regular masked array with NaN values
    verbose : boolean
        If true prints out extra information

    Returns
    -------
    ret_data : RadarData object
        An object containing the information read from the file

    """
    bfile = os.path.basename(file)

    if (bfile.startswith('PM') or bfile.startswith('PH') or
            bfile.startswith('PL') or bfile.startswith('MS') or
            bfile.startswith('MH') or bfile.startswith('ML')):
        # polar data from SITE (PH/PM/PL)
        if verbose:
            print("calling read_polar")
        ret = read_polar(
            file, moment=moment, physic_value=physic_value,
            masked_array=masked_array, verbose=verbose)
    else:
        # cartesian / CCS4 products
        if verbose:
            print("calling read_product")
        ret = read_product(
            file, physic_value=physic_value, masked_array=masked_array,
            verbose=verbose)

    return ret



"""
pyart.util.radar_utils
======================

Functions for working radar instances.

.. autosummary::
    :toctree: generated/

    is_vpt
    to_vpt
    join_radar
    join_spectra
    cut_radar
    cut_radar_spectra
    radar_from_spectra
    interpol_spectra
    ma_broadcast_to

"""

import copy
from warnings import warn

import numpy as np
from scipy.interpolate import interp1d
from netCDF4 import date2num

from ..config import get_fillvalue
from ..core import Radar
from . import datetime_utils


def is_vpt(radar, offset=0.5):
    """
    Determine if a Radar appears to be a vertical pointing scan.

    This function only verifies that the object is a vertical pointing scan,
    use the :py:func:`to_vpt` function to convert the radar to a vpt scan
    if this function returns True.

    Parameters
    ----------
    radar : Radar
        Radar object to determine if.
    offset : float, optional
        Maximum offset of the elevation from 90 degrees to still consider
        to be vertically pointing.

    Returns
    -------
    flag : bool
        True if the radar appear to be verticle pointing, False if not.

    """
    # check that the elevation is within offset of 90 degrees.
    elev = radar.elevation['data']
    return np.all((elev < 90.0 + offset) & (elev > 90.0 - offset))


def to_vpt(radar, single_scan=True):
    """
    Convert an existing Radar object to represent a vertical pointing scan.

    This function does not verify that the Radar object contains a vertical
    pointing scan. To perform such a check use :py:func:`is_vpt`.

    Parameters
    ----------
    radar : Radar
        Mislabeled vertical pointing scan Radar object to convert to be
        properly labeled. This object is converted in place, no copy of
        the existing data is made.
    single_scan : bool, optional
        True to convert the volume to a single scan, any azimuth angle data
        is lost. False will convert the scan to contain the same number of
        scans as rays, azimuth angles are retained.

    """
    if single_scan:
        nsweeps = 1
        radar.azimuth['data'][:] = 0.0
        seri = np.array([radar.nrays - 1], dtype='int32')
        radar.sweep_end_ray_index['data'] = seri
    else:
        nsweeps = radar.nrays
        # radar.azimuth not adjusted
        radar.sweep_end_ray_index['data'] = np.arange(nsweeps, dtype='int32')

    radar.scan_type = 'vpt'
    radar.nsweeps = nsweeps
    radar.target_scan_rate = None       # no scanning
    radar.elevation['data'][:] = 90.0

    radar.sweep_number['data'] = np.arange(nsweeps, dtype='int32')
    radar.sweep_mode['data'] = np.array(['vertical_pointing'] * nsweeps)
    radar.fixed_angle['data'] = np.ones(nsweeps, dtype='float32') * 90.0
    radar.sweep_start_ray_index['data'] = np.arange(nsweeps, dtype='int32')

    if radar.instrument_parameters is not None:
        for key in ['prt_mode', 'follow_mode', 'polarization_mode']:
            if key in radar.instrument_parameters:
                ip_dic = radar.instrument_parameters[key]
                ip_dic['data'] = np.array([ip_dic['data'][0]] * nsweeps)

    # Attributes that do not need any changes
    # radar.altitude
    # radar.altitude_agl
    # radar.latitude
    # radar.longitude

    # radar.range
    # radar.ngates
    # radar.nrays

    # radar.metadata
    # radar.radar_calibration

    # radar.time
    # radar.fields
    # radar.antenna_transition
    # radar.scan_rate


def join_radar(radar1, radar2):
    """
    Combine two radar instances into one.

    Parameters
    ----------
    radar1 : Radar
        Radar object.
    radar2 : Radar
        Radar object.

    """
    # must have same gate spacing
    new_radar = copy.deepcopy(radar1)
    new_radar.azimuth['data'] = np.append(
        radar1.azimuth['data'], radar2.azimuth['data'])
    new_radar.elevation['data'] = np.append(
        radar1.elevation['data'], radar2.elevation['data'])
    new_radar.fixed_angle['data'] = np.append(
        radar1.fixed_angle['data'], radar2.fixed_angle['data'])
    new_radar.sweep_number['data'] = np.append(
        radar1.sweep_number['data'], radar2.sweep_number['data'])
    new_radar.sweep_start_ray_index['data'] = np.append(
        radar1.sweep_start_ray_index['data'],
        radar2.sweep_start_ray_index['data'] + radar1.nrays)
    new_radar.sweep_end_ray_index['data'] = np.append(
        radar1.sweep_end_ray_index['data'],
        radar2.sweep_end_ray_index['data'] + radar1.nrays)
    new_radar.nsweeps += radar2.nsweeps
    new_radar.sweep_mode['data'] = np.append(
        radar1.sweep_mode['data'], radar2.sweep_mode['data'])
    if ((radar1.rays_are_indexed is not None) and
            (radar2.rays_are_indexed is not None)):
        new_radar.rays_are_indexed['data'] = np.append(
            radar1.rays_are_indexed['data'],
            radar2.rays_are_indexed['data'])
    else:
        new_radar.rays_are_indexed = None

    if new_radar.instrument_parameters is not None:
        if 'nyquist_velocity' in new_radar.instrument_parameters:
            new_radar.instrument_parameters['nyquist_velocity']['data'] = (
                np.append(
                    radar1.instrument_parameters['nyquist_velocity']['data'],
                    radar2.instrument_parameters['nyquist_velocity']['data']))
        if 'pulse_width' in new_radar.instrument_parameters:
            new_radar.instrument_parameters['pulse_width']['data'] = (
                np.append(
                    radar1.instrument_parameters['pulse_width']['data'],
                    radar2.instrument_parameters['pulse_width']['data']))
        if 'number_of_pulses' in new_radar.instrument_parameters:
            new_radar.instrument_parameters['number_of_pulses']['data'] = (
                np.append(
                    radar1.instrument_parameters['number_of_pulses']['data'],
                    radar2.instrument_parameters['number_of_pulses']['data']))

    if ((radar1.ray_angle_res is not None) and
            (radar2.ray_angle_res is not None)):
        new_radar.ray_angle_res['data'] = np.append(
            radar1.ray_angle_res['data'],
            radar2.ray_angle_res['data'])
    else:
        new_radar.ray_angle_res = None

    if len(radar1.range['data']) >= len(radar2.range['data']):
        new_radar.range['data'] = radar1.range['data']
    else:
        new_radar.range['data'] = radar2.range['data']
    new_radar.ngates = len(new_radar.range['data'])

    # to combine times we need to reference them to a standard
    # for this we'll use epoch time
    r1num = datetime_utils.datetimes_from_radar(radar1, epoch=True)
    r2num = datetime_utils.datetimes_from_radar(radar2, epoch=True)
    new_radar.time['data'] = date2num(
        np.append(r1num, r2num), datetime_utils.EPOCH_UNITS)
    new_radar.time['units'] = datetime_utils.EPOCH_UNITS
    new_radar.nrays = len(new_radar.time['data'])

    fields_to_remove = []
    for var in new_radar.fields.keys():
        # if the field is present in both radars combine both fields
        # otherwise remove it from new radar
        if var in radar1.fields and var in radar2.fields:
            sh1 = radar1.fields[var]['data'].shape
            sh2 = radar2.fields[var]['data'].shape
            new_field_shape = (sh1[0] + sh2[0], max(sh1[1], sh2[1]))
            new_field = np.ma.masked_all(new_field_shape)
            new_field.set_fill_value(get_fillvalue())
            new_field[0:sh1[0], 0:sh1[1]] = radar1.fields[var]['data']
            new_field[sh1[0]:, 0:sh2[1]] = radar2.fields[var]['data']
            new_radar.fields[var]['data'] = new_field
        else:
            warn("Field "+var+" not present in both radars")
            fields_to_remove.append(var)

    if fields_to_remove:
        for field_name in fields_to_remove:
            new_radar.fields.pop(field_name, None)

    # radar locations
    # TODO moving platforms - any more?
    if (len(radar1.latitude['data']) == 1 &
            len(radar2.latitude['data']) == 1 &
            len(radar1.longitude['data']) == 1 &
            len(radar2.longitude['data']) == 1 &
            len(radar1.altitude['data']) == 1 &
            len(radar2.altitude['data']) == 1):

        lat1 = float(radar1.latitude['data'])
        lon1 = float(radar1.longitude['data'])
        alt1 = float(radar1.altitude['data'])
        lat2 = float(radar2.latitude['data'])
        lon2 = float(radar2.longitude['data'])
        alt2 = float(radar2.altitude['data'])

        if (lat1 != lat2) or (lon1 != lon2) or (alt1 != alt2):
            ones1 = np.ones(len(radar1.time['data']), dtype='float32')
            ones2 = np.ones(len(radar2.time['data']), dtype='float32')
            new_radar.latitude['data'] = np.append(ones1 * lat1, ones2 * lat2)
            new_radar.longitude['data'] = np.append(ones1 * lon1, ones2 * lon2)
            new_radar.latitude['data'] = np.append(ones1 * alt1, ones2 * alt2)
        else:
            new_radar.latitude['data'] = radar1.latitude['data']
            new_radar.longitude['data'] = radar1.longitude['data']
            new_radar.altitude['data'] = radar1.altitude['data']

    else:
        new_radar.latitude['data'] = np.append(radar1.latitude['data'],
                                               radar2.latitude['data'])
        new_radar.longitude['data'] = np.append(radar1.longitude['data'],
                                                radar2.longitude['data'])
        new_radar.altitude['data'] = np.append(radar1.altitude['data'],
                                               radar2.altitude['data'])

    return new_radar


def join_spectra(spectra1, spectra2):
    """
    Combine two spectra instances into one.

    Parameters
    ----------
    spectra1 : spectra
        spectra object.
    spectra2 : spectra
        spectra object.

    """
    # must have same gate spacing
    new_spectra = copy.deepcopy(spectra1)
    new_spectra.azimuth['data'] = np.append(
        spectra1.azimuth['data'], spectra2.azimuth['data'])
    new_spectra.elevation['data'] = np.append(
        spectra1.elevation['data'], spectra2.elevation['data'])
    new_spectra.fixed_angle['data'] = np.append(
        spectra1.fixed_angle['data'], spectra2.fixed_angle['data'])
    new_spectra.sweep_number['data'] = np.append(
        spectra1.sweep_number['data'], spectra2.sweep_number['data'])
    new_spectra.sweep_start_ray_index['data'] = np.append(
        spectra1.sweep_start_ray_index['data'],
        spectra2.sweep_start_ray_index['data'] + spectra1.nrays)
    new_spectra.sweep_end_ray_index['data'] = np.append(
        spectra1.sweep_end_ray_index['data'],
        spectra2.sweep_end_ray_index['data'] + spectra1.nrays)
    new_spectra.nsweeps += spectra2.nsweeps
    new_spectra.sweep_mode['data'] = np.append(
        spectra1.sweep_mode['data'], spectra2.sweep_mode['data'])
    if ((spectra1.rays_are_indexed is not None) and
            (spectra2.rays_are_indexed is not None)):
        new_spectra.rays_are_indexed['data'] = np.append(
            spectra1.rays_are_indexed['data'],
            spectra2.rays_are_indexed['data'])
    else:
        new_spectra.rays_are_indexed = None

    if new_spectra.instrument_parameters is not None:
        if 'nyquist_velocity' in new_spectra.instrument_parameters:
            new_spectra.instrument_parameters['nyquist_velocity']['data'] = (
                np.append(
                    spectra1.instrument_parameters['nyquist_velocity']['data'],
                    spectra2.instrument_parameters['nyquist_velocity']['data']))
        if 'pulse_width' in new_spectra.instrument_parameters:
            new_spectra.instrument_parameters['pulse_width']['data'] = (
                np.append(
                    spectra1.instrument_parameters['pulse_width']['data'],
                    spectra2.instrument_parameters['pulse_width']['data']))
        if 'number_of_pulses' in new_spectra.instrument_parameters:
            new_spectra.instrument_parameters['number_of_pulses']['data'] = (
                np.append(
                    spectra1.instrument_parameters['number_of_pulses']['data'],
                    spectra2.instrument_parameters['number_of_pulses']['data']))

    if ((spectra1.ray_angle_res is not None) and
            (spectra2.ray_angle_res is not None)):
        new_spectra.ray_angle_res['data'] = np.append(
            spectra1.ray_angle_res['data'],
            spectra2.ray_angle_res['data'])
    else:
        new_spectra.ray_angle_res = None

    if len(spectra1.range['data']) >= len(spectra2.range['data']):
        new_spectra.range['data'] = spectra1.range['data']
    else:
        new_spectra.range['data'] = spectra2.range['data']
    new_spectra.ngates = len(new_spectra.range['data'])

    # Combine Doppler parameters
    if spectra1.npulses_max != spectra2.npulses_max:
        new_spectra.npulses_max = max(
            spectra1.npulses_max, spectra2.npulses_max)

    new_spectra.npulses['data'] = np.append(
        spectra1.npulses['data'], spectra2.npulses['data'])

    if (spectra1.Doppler_velocity is not None and
            spectra2.Doppler_velocity is not None):
        sh1 = spectra1.Doppler_velocity['data'].shape
        sh2 = spectra2.Doppler_velocity['data'].shape
        new_field = np.ma.masked_all((sh1[0] + sh2[0], max(sh1[1], sh2[1])))
        new_field[0:sh1[0], 0:sh1[1]] = spectra1.Doppler_velocity['data']
        new_field[sh1[0]:, 0:sh2[1]] = spectra2.Doppler_velocity['data']
        new_spectra.Doppler_velocity['data'] = new_field
    else:
        new_spectra.Doppler_velocity = None

    if (spectra1.Doppler_frequency is not None and
            spectra2.Doppler_frequency is not None):
        sh1 = spectra1.Doppler_frequency['data'].shape
        sh2 = spectra2.Doppler_frequency['data'].shape
        new_field = np.ma.masked_all((sh1[0] + sh2[0], max(sh1[1], sh2[1])))
        new_field[0:sh1[0], 0:sh1[1]] = spectra1.Doppler_frequency['data']
        new_field[sh1[0]:, 0:sh2[1]] = spectra2.Doppler_frequency['data']
        new_spectra.Doppler_frequency['data'] = new_field
    else:
        new_spectra.Doppler_frequency = None

    # to combine times we need to reference them to a standard
    # for this we'll use epoch time
    r1num = datetime_utils.datetimes_from_radar(spectra1, epoch=True)
    r2num = datetime_utils.datetimes_from_radar(spectra2, epoch=True)
    new_spectra.time['data'] = date2num(
        np.append(r1num, r2num), datetime_utils.EPOCH_UNITS)
    new_spectra.time['units'] = datetime_utils.EPOCH_UNITS
    new_spectra.nrays = len(new_spectra.time['data'])

    fields_to_remove = []
    for var in new_spectra.fields.keys():
        # if the field is present in both spectras combine both fields
        # otherwise remove it from new spectra
        if var in spectra1.fields and var in spectra2.fields:
            sh1 = spectra1.fields[var]['data'].shape
            sh2 = spectra2.fields[var]['data'].shape
            new_field = np.ma.masked_all(
                (sh1[0] + sh2[0], max(sh1[1], sh2[1]), max(sh1[2], sh2[2])))
            new_field.set_fill_value(get_fillvalue())
            new_field[0:sh1[0], 0:sh1[1], 0:sh1[2]] = (
                spectra1.fields[var]['data'])
            new_field[sh1[0]:, 0:sh2[1], 0:sh2[2]] = (
                spectra2.fields[var]['data'])
            new_spectra.fields[var]['data'] = new_field
        else:
            warn("Field "+var+" not present in both spectras")
            fields_to_remove.append(var)

    if fields_to_remove:
        for field_name in fields_to_remove:
            new_spectra.fields.pop(field_name, None)

    # spectra locations
    # TODO moving platforms - any more?
    if (len(spectra1.latitude['data']) == 1 &
            len(spectra2.latitude['data']) == 1 &
            len(spectra1.longitude['data']) == 1 &
            len(spectra2.longitude['data']) == 1 &
            len(spectra1.altitude['data']) == 1 &
            len(spectra2.altitude['data']) == 1):

        lat1 = float(spectra1.latitude['data'])
        lon1 = float(spectra1.longitude['data'])
        alt1 = float(spectra1.altitude['data'])
        lat2 = float(spectra2.latitude['data'])
        lon2 = float(spectra2.longitude['data'])
        alt2 = float(spectra2.altitude['data'])

        if (lat1 != lat2) or (lon1 != lon2) or (alt1 != alt2):
            ones1 = np.ones(len(spectra1.time['data']), dtype='float32')
            ones2 = np.ones(len(spectra2.time['data']), dtype='float32')
            new_spectra.latitude['data'] = np.append(ones1 * lat1, ones2 * lat2)
            new_spectra.longitude['data'] = np.append(ones1 * lon1, ones2 * lon2)
            new_spectra.latitude['data'] = np.append(ones1 * alt1, ones2 * alt2)
        else:
            new_spectra.latitude['data'] = spectra1.latitude['data']
            new_spectra.longitude['data'] = spectra1.longitude['data']
            new_spectra.altitude['data'] = spectra1.altitude['data']

    else:
        new_spectra.latitude['data'] = np.append(
            spectra1.latitude['data'], spectra2.latitude['data'])
        new_spectra.longitude['data'] = np.append(
            spectra1.longitude['data'], spectra2.longitude['data'])
        new_spectra.altitude['data'] = np.append(
            spectra1.altitude['data'], spectra2.altitude['data'])

    return new_spectra


def cut_radar(radar, field_names, rng_min=None, rng_max=None, ele_min=None,
              ele_max=None, azi_min=None, azi_max=None):
    """
    Cuts the radar object into new dimensions

    Parameters
    ----------
    radar : radar object
        The radar object containing the data
    field_names : str or None
        The fields to keep in the new radar
    rng_min, rng_max : float
        The range limits [m]. If None the entire coverage of the radar is
        going to be used
    ele_min, ele_max, azi_min, azi_max : float or None
        The limits of the grid [deg]. If None the limits will be the limits
        of the radar volume

    Returns
    -------
    radar : radar object
        The radar object containing only the desired data

    """
    radar_aux = copy.deepcopy(radar)

    if (rng_min is None and rng_max is None and ele_min is None and
            ele_max is None and azi_min is None and azi_max is None):
        return radar_aux

    if rng_min is None:
        rng_min = 0.
    if rng_max is None:
        rng_max = np.max(radar_aux.range['data'])

    ind_rng = np.where(np.logical_and(
        radar_aux.range['data'] >= rng_min, radar_aux.range['data'] <= rng_max))[0]

    if ind_rng.size == 0:
        warn('No range bins between '+str(rng_min)+' and '+str(rng_max)+' m')
        return None

    # Determine angle limits
    if radar_aux.scan_type == 'ppi':
        if ele_min is None:
            ele_min = np.min(radar_aux.fixed_angle['data'])
        if ele_max is None:
            ele_max = np.max(radar_aux.fixed_angle['data'])
        if azi_min is None:
            azi_min = np.min(radar_aux.azimuth['data'])
        if azi_max is None:
            azi_max = np.max(radar_aux.azimuth['data'])
    else:
        if ele_min is None:
            ele_min = np.min(radar_aux.elevation['data'])
        if ele_max is None:
            ele_max = np.max(radar_aux.elevation['data'])
        if azi_min is None:
            azi_min = np.min(radar_aux.fixed_angle['data'])
        if azi_max is None:
            azi_max = np.max(radar_aux.fixed_angle['data'])

    if radar_aux.scan_type == 'ppi':
        # Get radar elevation angles within limits
        ele_vec = np.sort(radar_aux.fixed_angle['data'])
        ele_vec = ele_vec[
            np.logical_and(ele_vec >= ele_min, ele_vec <= ele_max)]
        if ele_vec.size == 0:
            warn('No elevation angles between '+str(ele_min)+' and ' +
                 str(ele_max))
            return None

        # get sweeps corresponding to the desired elevation angles
        ind_sweeps = []
        for ele in ele_vec:
            ind_sweeps.append(
                np.where(radar_aux.fixed_angle['data'] == ele)[0][0])
        radar_aux = radar_aux.extract_sweeps(ind_sweeps)

        # Get indices of rays within limits
        if azi_min < azi_max:
            ind_rays = np.where(np.logical_and(
                radar_aux.azimuth['data'] >= azi_min,
                radar_aux.azimuth['data'] <= azi_max))[0]
        else:
            ind_rays = np.where(np.logical_or(
                radar_aux.azimuth['data'] >= azi_min,
                radar_aux.azimuth['data'] <= azi_max))[0]

    else:
        # Get radar azimuth angles within limits
        azi_vec = radar_aux.fixed_angle['data']
        if azi_min < azi_max:
            azi_vec = np.sort(azi_vec[
                np.logical_and(azi_vec >= azi_min, azi_vec <= azi_max)])
        else:
            azi_vec = azi_vec[
                np.logical_or(azi_vec >= azi_min, azi_vec <= azi_max)]
        if azi_vec.size == 0:
            warn('No azimuth angles between '+str(azi_min)+' and ' +
                 str(azi_max))
            return None

        # get sweeps corresponding to the desired azimuth angles
        ind_sweeps = []
        for azi in azi_vec:
            ind_sweeps.append(
                np.where(radar_aux.fixed_angle['data'] == azi)[0][0])
        radar_aux = radar_aux.extract_sweeps(ind_sweeps)

        # Get indices of rays within limits
        ind_rays = np.where(np.logical_and(
            radar_aux.elevation['data'] >= ele_min,
            radar_aux.elevation['data'] <= ele_max))[0]

    # get new sweep start index and stop index
    sweep_start_inds = copy.deepcopy(radar_aux.sweep_start_ray_index['data'])
    sweep_end_inds = copy.deepcopy(radar_aux.sweep_end_ray_index['data'])

    nrays = 0
    for j in range(radar_aux.nsweeps):
        # get azimuth indices for this elevation
        rays_in_sweep = np.size(
            ind_rays[np.logical_and(ind_rays >= sweep_start_inds[j],
                                    ind_rays <= sweep_end_inds[j])])
        radar_aux.rays_per_sweep['data'][j] = rays_in_sweep
        if j == 0:
            radar_aux.sweep_start_ray_index['data'][j] = 0
        else:
            radar_aux.sweep_start_ray_index['data'][j] = int(
                radar_aux.sweep_end_ray_index['data'][j-1]+1)
        radar_aux.sweep_end_ray_index['data'][j] = (
            radar_aux.sweep_start_ray_index['data'][j]+rays_in_sweep-1)
        nrays += rays_in_sweep

    # Update metadata
    radar_aux.range['data'] = radar_aux.range['data'][ind_rng]
    radar_aux.time['data'] = radar_aux.time['data'][ind_rays]
    radar_aux.azimuth['data'] = radar_aux.azimuth['data'][ind_rays]
    radar_aux.elevation['data'] = radar_aux.elevation['data'][ind_rays]
    radar_aux.init_gate_x_y_z()
    radar_aux.init_gate_longitude_latitude()
    radar_aux.init_gate_altitude()
    radar_aux.nrays = nrays
    radar_aux.ngates = ind_rng.size

    if radar_aux.instrument_parameters is not None:
        if 'nyquist_velocity' in radar_aux.instrument_parameters:
            radar_aux.instrument_parameters['nyquist_velocity']['data'] = (
                radar_aux.instrument_parameters['nyquist_velocity']['data'][
                    ind_rays])
        if 'pulse_width' in radar_aux.instrument_parameters:
            radar_aux.instrument_parameters['pulse_width']['data'] = (
                radar_aux.instrument_parameters['pulse_width']['data'][
                    ind_rays])
        if 'number_of_pulses' in radar_aux.instrument_parameters:
            radar_aux.instrument_parameters['number_of_pulses']['data'] = (
                radar_aux.instrument_parameters['number_of_pulses']['data'][
                    ind_rays])

    # Get new fields
    if field_names is None:
        radar_aux.fields = dict()
    else:
        fields_aux = copy.deepcopy(radar_aux.fields)
        radar_aux.fields = dict()
        for field_name in field_names:
            if field_name not in fields_aux:
                warn('Field '+field_name+' not available')
                continue

            fields_aux[field_name]['data'] = (
                fields_aux[field_name]['data'][:, ind_rng])
            fields_aux[field_name]['data'] = (
                fields_aux[field_name]['data'][ind_rays, :])
            radar_aux.add_field(field_name, fields_aux[field_name])

    return radar_aux


def cut_radar_spectra(radar, field_names, rng_min=None, rng_max=None,
                      ele_min=None, ele_max=None, azi_min=None, azi_max=None):
    """
    Cuts the radar spectra object into new dimensions

    Parameters
    ----------
    radar : radar object
        The radar object containing the data
    field_names : str or None
        The fields to keep in the new radar
    rng_min, rng_max : float
        The range limits [m]. If None the entire coverage of the radar is
        going to be used
    ele_min, ele_max, azi_min, azi_max : float or None
        The limits of the grid [deg]. If None the limits will be the limits
        of the radar volume

    Returns
    -------
    radar : radar object
        The radar object containing only the desired data

    """
    radar_aux = copy.deepcopy(radar)

    if (rng_min is None and rng_max is None and ele_min is None and
            ele_max is None and azi_min is None and azi_max is None):
        return radar_aux

    if rng_min is None:
        rng_min = 0.
    if rng_max is None:
        rng_max = np.max(radar_aux.range['data'])

    ind_rng = np.where(np.logical_and(
        radar_aux.range['data'] >= rng_min, radar_aux.range['data'] <= rng_max))[0]

    if ind_rng.size == 0:
        warn('No range bins between '+str(rng_min)+' and '+str(rng_max)+' m')
        return None

    # Determine angle limits
    if radar_aux.scan_type == 'ppi':
        if ele_min is None:
            ele_min = np.min(radar_aux.fixed_angle['data'])
        if ele_max is None:
            ele_max = np.max(radar_aux.fixed_angle['data'])
        if azi_min is None:
            azi_min = np.min(radar_aux.azimuth['data'])
        if azi_max is None:
            azi_max = np.max(radar_aux.azimuth['data'])
    else:
        if ele_min is None:
            ele_min = np.min(radar_aux.elevation['data'])
        if ele_max is None:
            ele_max = np.max(radar_aux.elevation['data'])
        if azi_min is None:
            azi_min = np.min(radar_aux.fixed_angle['data'])
        if azi_max is None:
            azi_max = np.max(radar_aux.fixed_angle['data'])

    if radar_aux.scan_type == 'ppi':
        # Get radar elevation angles within limits
        ele_vec = np.sort(radar_aux.fixed_angle['data'])
        ele_vec = ele_vec[
            np.logical_and(ele_vec >= ele_min, ele_vec <= ele_max)]
        if ele_vec.size == 0:
            warn('No elevation angles between '+str(ele_min)+' and ' +
                 str(ele_max))
            return None

        # get sweeps corresponding to the desired elevation angles
        ind_sweeps = []
        for ele in ele_vec:
            ind_sweeps.append(
                np.where(radar_aux.fixed_angle['data'] == ele)[0][0])
        radar_aux = radar_aux.extract_sweeps(ind_sweeps)

        # Get indices of rays within limits
        if azi_min < azi_max:
            ind_rays = np.where(np.logical_and(
                radar_aux.azimuth['data'] >= azi_min,
                radar_aux.azimuth['data'] <= azi_max))[0]
        else:
            ind_rays = np.where(np.logical_or(
                radar_aux.azimuth['data'] >= azi_min,
                radar_aux.azimuth['data'] <= azi_max))[0]

    else:
        # Get radar azimuth angles within limits
        azi_vec = radar_aux.fixed_angle['data']
        if azi_min < azi_max:
            azi_vec = np.sort(azi_vec[
                np.logical_and(azi_vec >= azi_min, azi_vec <= azi_max)])
        else:
            azi_vec = azi_vec[
                np.logical_or(azi_vec >= azi_min, azi_vec <= azi_max)]
        if azi_vec.size == 0:
            warn('No azimuth angles between '+str(azi_min)+' and ' +
                 str(azi_max))
            return None

        # get sweeps corresponding to the desired azimuth angles
        ind_sweeps = []
        for azi in azi_vec:
            ind_sweeps.append(
                np.where(radar_aux.fixed_angle['data'] == azi)[0][0])
        radar_aux = radar_aux.extract_sweeps(ind_sweeps)

        # Get indices of rays within limits
        ind_rays = np.where(np.logical_and(
            radar_aux.elevation['data'] >= ele_min,
            radar_aux.elevation['data'] <= ele_max))[0]

    # get new sweep start index and stop index
    sweep_start_inds = copy.deepcopy(radar_aux.sweep_start_ray_index['data'])
    sweep_end_inds = copy.deepcopy(radar_aux.sweep_end_ray_index['data'])

    nrays = 0
    for j in range(radar_aux.nsweeps):
        # get azimuth indices for this elevation
        rays_in_sweep = np.size(
            ind_rays[np.logical_and(ind_rays >= sweep_start_inds[j],
                                    ind_rays <= sweep_end_inds[j])])
        radar_aux.rays_per_sweep['data'][j] = rays_in_sweep
        if j == 0:
            radar_aux.sweep_start_ray_index['data'][j] = 0
        else:
            radar_aux.sweep_start_ray_index['data'][j] = int(
                radar_aux.sweep_end_ray_index['data'][j-1]+1)
        radar_aux.sweep_end_ray_index['data'][j] = (
            radar_aux.sweep_start_ray_index['data'][j]+rays_in_sweep-1)
        nrays += rays_in_sweep

    # Update metadata
    radar_aux.range['data'] = radar_aux.range['data'][ind_rng]
    radar_aux.time['data'] = radar_aux.time['data'][ind_rays]
    radar_aux.azimuth['data'] = radar_aux.azimuth['data'][ind_rays]
    radar_aux.elevation['data'] = radar_aux.elevation['data'][ind_rays]
    radar_aux.init_gate_x_y_z()
    radar_aux.init_gate_longitude_latitude()
    radar_aux.init_gate_altitude()
    radar_aux.nrays = nrays
    radar_aux.ngates = ind_rng.size
    if radar_aux.Doppler_velocity is not None:
        radar_aux.Doppler_velocity['data'] = (
            radar_aux.Doppler_velocity['data'][ind_rays, :])
    if radar_aux.Doppler_frequency is not None:
        radar_aux.Doppler_frequency['data'] = (
            radar_aux.Doppler_frequency['data'][ind_rays, :])
    if radar_aux.instrument_parameters is not None:
        if 'nyquist_velocity' in radar_aux.instrument_parameters:
            radar_aux.instrument_parameters['nyquist_velocity']['data'] = (
                radar_aux.instrument_parameters['nyquist_velocity']['data'][
                    ind_rays])
        if 'pulse_width' in radar_aux.instrument_parameters:
            radar_aux.instrument_parameters['pulse_width']['data'] = (
                radar_aux.instrument_parameters['pulse_width']['data'][
                    ind_rays])
        if 'number_of_pulses' in radar_aux.instrument_parameters:
            radar_aux.instrument_parameters['number_of_pulses']['data'] = (
                radar_aux.instrument_parameters['number_of_pulses']['data'][
                    ind_rays])

    # Get new fields
    if field_names is None:
        radar_aux.fields = dict()
    else:
        fields_aux = copy.deepcopy(radar_aux.fields)
        radar_aux.fields = dict()
        for field_name in field_names:
            if field_name not in fields_aux:
                warn('Field '+field_name+' not available')
                continue

            fields_aux[field_name]['data'] = (
                fields_aux[field_name]['data'][:, ind_rng, :])
            fields_aux[field_name]['data'] = (
                fields_aux[field_name]['data'][ind_rays, :, :])
            radar_aux.add_field(field_name, fields_aux[field_name])

    return radar_aux


def radar_from_spectra(psr):
    """
    obtain a Radar object from a RadarSpectra object

    Parameters
    ----------
    psr : RadarSpectra object
        The reference object

    Returns
    -------
    radar : radar object
        The new radar object

    """
    return Radar(
        psr.time, psr.range, dict(), psr.metadata, psr.scan_type,
        psr.latitude, psr.longitude, psr.altitude, psr.sweep_number,
        psr.sweep_mode, psr.fixed_angle, psr.sweep_start_ray_index,
        psr.sweep_end_ray_index, psr.azimuth, psr.elevation,
        rays_are_indexed=psr.rays_are_indexed,
        ray_angle_res=psr.ray_angle_res,
        instrument_parameters=psr.instrument_parameters,
        radar_calibration=psr.radar_calibration)


def interpol_spectra(psr, kind='linear', fill_value=0.):
    """
    Interpolates the spectra so that it has a uniform grid

    Parameters
    ----------
    psr : RadarSpectra object
        The original spectra

    Returns
    -------
    psr_interp : RadarSpectra object
        The interpolated spectra

    """
    if psr.Doppler_velocity is not None:
        # check if all rays have the same Doppler velocity field
        if np.ma.all(np.ma.all(
                psr.Doppler_velocity['data'] == psr.Doppler_velocity['data'][0, :],
                axis=1)):
            return copy.deepcopy(psr)

        xaxis = psr.Doppler_velocity['data'][0, :].compressed()
        units = 'm/s'
    elif psr.Doppler_frequency is not None:
        # check if all rays have the same Doppler velocity field
        if np.ma.all(np.ma.all(
                psr.Doppler_frequency['data'] == psr.Doppler_frequency['data'][0, :],
                axis=1)):
            return copy.deepcopy(psr)

        xaxis = psr.Doppler_frequency['data'][0, :].compressed()
        units = 'Hz'
    else:
        warn('Unable to interpolate spectra. No Doppler velocity or Doppler'
             ' frequency available')
        return None
    npulses = xaxis.size

    psr_interp = copy.deepcopy(psr)

    # allocate memory for output fields
    for field_name in psr_interp.fields.keys():
        psr_interp.fields[field_name]['data'] = np.ma.masked_all(
            (psr_interp.nrays, psr_interp.ngates, npulses),
            dtype=psr.fields[field_name]['data'].dtype)

    for ray in range(psr.nrays):
        if units == 'm/s':
            xaxis_aux = psr.Doppler_velocity['data'][ray, :].compressed()
        else:
            xaxis_aux = psr.Doppler_frequency['data'][ray, :].compressed()

        for field_name in psr_interp.fields.keys():
            intp_func = interp1d(
                xaxis_aux, psr.fields[field_name]['data'][
                    ray, :, 0:xaxis_aux.size].filled(fill_value), kind=kind,
                fill_value='extrapolate', assume_sorted=True)

            psr_interp.fields[field_name]['data'][ray, :, 0:npulses] = (
                np.ma.masked_invalid(intp_func(xaxis)))

    psr_interp.npulses_max = npulses
    psr_interp.npulses['data'] = (
        np.zeros(psr_interp.nrays, dtype=np.int)+npulses)
    if psr_interp.Doppler_velocity is not None:
        xaxis = np.ma.expand_dims(
            psr.Doppler_velocity['data'][0, :].compressed(), axis=0)
        psr_interp.Doppler_velocity['data'] = ma_broadcast_to(
            xaxis, (psr_interp.nrays, psr_interp.npulses_max))
        psr_interp.Doppler_velocity['data'] = np.ma.array(
            psr_interp.Doppler_velocity['data'])

    if psr_interp.Doppler_frequency is not None:
        xaxis = np.ma.expand_dims(
            psr.Doppler_frequency['data'][0, :].compressed(), axis=0)
        psr_interp.Doppler_frequency['data'] = ma_broadcast_to(
            xaxis, (psr_interp.nrays, psr_interp.npulses_max))
        psr_interp.Doppler_frequency['data'] = np.ma.array(
            psr_interp.Doppler_frequency['data'])

    return psr_interp


def ma_broadcast_to(array, tup):
    """
    Is used to guarantee that a masked array can be broadcasted without loosing the mask

    Parameters
    ----------
    array : Numpy masked array or normal array
    tup : shape as tuple

    Returns
    -------
    broadcasted_array
        The broadcasted numpy array including its mask if available
        otherwise only the broadcasted array is returned

    """
    broadcasted_array = np.broadcast_to(array, tup)

    if np.ma.is_masked(array):
        initial_mask = np.ma.getmask(array)
        initial_fill_value = array.fill_value
        broadcasted_mask = np.broadcast_to(initial_mask, tup)
        return np.ma.array(broadcasted_array, mask=broadcasted_mask, fill_value=initial_fill_value)

    return broadcasted_array



"""
pyart.bridge.wradlib
====================

Py-ART methods linking to wradlib functions, http://wradlib.org/

.. autosummary::
    :toctree: generated/

     texture_of_complex_phase

"""

try:
    import wradlib
    _WRADLIB_AVAILABLE = True
except ImportError:
    _WRADLIB_AVAILABLE = False
import numpy as np

from ..config import get_metadata, get_field_name
from ..exceptions import MissingOptionalDependency


def texture_of_complex_phase(radar, phidp_field=None,
                             phidp_texture_field=None):
    """
    Calculate the texture of the differential phase field.

    Calculate the texture of the real part of the complex differential
    phase field

    Parameters
    ----------
    radar : Radar
        Radar object from which to .
    phidp_field : str, optional
        Name of field in radar which contains the differential phase shift.
        None will use the default field name in the Py-ART configuration file.
    phidp_texture_field : str, optional
        Name to use for the differential phase texture field metadata.
        None will use the default field name in the Py-ART configuration file.

    Returns
    -------
    texture_field : dict
        Field dictionary containing the texture of the real part
        of the complex differential phase.

    References
    ----------
    Gourley, J. J., P. Tabary, and J. Parent du Chatelet,
    A fuzzy logic algorithm for the separation of precipitating from
    nonprecipitating echoes using polarimetric radar observations,
    Journal of Atmospheric and Oceanic Technology 24 (8), 1439-1451

    """
    # check that wradlib is available
    if not _WRADLIB_AVAILABLE:
        raise MissingOptionalDependency(
            "wradlib is required to use texture_of_complex_phase but is "
            + "not installed")

    # parse field names
    if phidp_field is None:
        phidp_field = get_field_name('differential_phase')

    if phidp_texture_field is None:
        phidp_texture_field = get_field_name('differential_phase')

    # Grab the phase data
    phidp = radar.fields[phidp_field]['data']

    # convert to complex number
    complex_phase = np.exp(1j*(phidp*np.pi/180.0))

    # calculate texture using wradlib
    w_texture_complex = wradlib.dp.texture(
        (np.real(complex_phase) + 1.0) * 180)

    texture_field = get_metadata(phidp_texture_field)
    texture_field['data'] = w_texture_complex
    texture_field['standard_name'] = 'texture_of_differential_phase_hv'
    texture_field['long_name'] = 'Texture of differential phase'
    return texture_field



"""
pyart.io.nexrad_cdm
===================

Functions for accessing Common Data Model (CDM) NEXRAD Level 2 files.

.. autosummary::
    :toctree: generated/

    read_nexrad_cdm
    _scan_info
    _populate_scan_dic
    _get_moment_data

"""

from datetime import datetime, timedelta
import os

import netCDF4
import numpy as np

from .nexrad_common import get_nexrad_location
from ..config import FileMetadata, get_fillvalue
from ..core.radar import Radar
from .common import make_time_unit_str, _test_arguments


def read_nexrad_cdm(filename, field_names=None, additional_metadata=None,
                    file_field_names=False, exclude_fields=None,
                    include_fields=None, station=None, **kwargs):
    """
    Read a Common Data Model (CDM) NEXRAD Level 2 file.

    Parameters
    ----------
    filename : str
        File name or URL of a Common Data Model (CDM) NEXRAD Level 2 file.
        File of in this format can be created using the NetCDF Java Library
        tools [1]_. A URL of a OPeNDAP file on the UCAR THREDDS Data
        Server [2]_ is also accepted the netCDF4 library has been compiled
        with OPeNDAP support.
    field_names : dict, optional
        Dictionary mapping NEXRAD moments to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        metadata configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included. A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the metadata configuration file will be used.
    file_field_names : bool, optional
        True to use the NEXRAD field names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    station : str
        Four letter ICAO name of the NEXRAD station used to determine the
        location in the returned radar object. This parameter is only
        used when the location is not contained in the file, which occur
        in older NEXRAD files. If the location is not provided in the file
        and this parameter is set to None the station name will be determined
        from the filename.

    Returns
    -------
    radar : Radar
        Radar object containing all moments and sweeps/cuts in the volume.
        Gates not collected are masked in the field data.

    References
    ----------
    .. [1] http://www.unidata.ucar.edu/software/netcdf-java/documentation.htm
    .. [2] http://thredds.ucar.edu/thredds/catalog.html

    """
    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    filemetadata = FileMetadata('nexrad_cdm', field_names,
                                additional_metadata, file_field_names,
                                exclude_fields, include_fields)

    # open the file
    dataset = netCDF4.Dataset(filename)
    dattrs = dataset.ncattrs()
    dvars = dataset.variables
    if 'cdm_data_type' not in dattrs or dataset.cdm_data_type != 'RADIAL':
        raise IOError('%s is not a valid CDM NetCDF file' % (filename))

    # determine the scan information
    scan_info = _scan_info(dvars)
    radials_per_scan = [max(s['nradials']) for s in scan_info]
    ngates_per_scan = [max(s['ngates']) for s in scan_info]
    ngates = max(ngates_per_scan)
    nrays = sum(radials_per_scan)
    nsweeps = len(scan_info)

    # extract data which changes depending on scan,
    # specifically time, azimuth, elevation and fixed angle data, as well as
    # the moment data.
    time_data = np.empty((nrays, ), dtype='float64')
    azim_data = np.empty((nrays, ), dtype='float32')
    elev_data = np.empty((nrays, ), dtype='float32')
    fixed_agl_data = np.empty((nsweeps, ), dtype='float32')
    fdata = {
        'Reflectivity':
        np.ma.masked_equal(np.ones((nrays, ngates), dtype='float32'), 1),
        'RadialVelocity':
        np.ma.masked_equal(np.ones((nrays, ngates), dtype='float32'), 1),
        'SpectrumWidth':
        np.ma.masked_equal(np.ones((nrays, ngates), dtype='float32'), 1),
        'DifferentialReflectivity':
        np.ma.masked_equal(np.ones((nrays, ngates), dtype='float32'), 1),
        'CorrelationCoefficient':
        np.ma.masked_equal(np.ones((nrays, ngates), dtype='float32'), 1),
        'DifferentialPhase':
        np.ma.masked_equal(np.ones((nrays, ngates), dtype='float32'), 1),
    }

    ray_i = 0
    for scan_index, scan_dic in enumerate(scan_info):

        var_index = scan_dic['index'][0]
        nradials = scan_dic['nradials'][0]

        time_var = scan_dic['time_vars'][0]
        azimuth_var = scan_dic['azimuth_vars'][0]
        elevation_var = scan_dic['elevation_vars'][0]
        nradials = scan_dic['nradials'][0]

        end = ray_i + nradials
        time_data[ray_i:end] = dvars[time_var][var_index][:nradials]
        azim_data[ray_i:end] = dvars[azimuth_var][var_index][:nradials]
        elev_data[ray_i:end] = dvars[elevation_var][var_index][:nradials]
        fixed_agl_data[scan_index] = np.mean(
            dvars[elevation_var][var_index][:nradials])

        for i, moment in enumerate(scan_dic['moments']):

            moment_index = scan_dic['index'][i]
            m_ngates = scan_dic['ngates'][i]
            m_nradials = scan_dic['nradials'][i]

            if moment.endswith('_HI'):
                fdata_name = moment[:-3]
            else:
                fdata_name = moment

            sweep = _get_moment_data(dvars[moment], moment_index, m_ngates)
            fdata[fdata_name][ray_i:ray_i + m_nradials, :m_ngates] = (
                sweep[:m_nradials, :m_ngates])

        ray_i += nradials

    # time
    time = filemetadata('time')
    first_time_var = scan_info[0]['time_vars'][0]
    time_start = datetime.strptime(dvars[first_time_var].units[-20:],
                                   "%Y-%m-%dT%H:%M:%SZ")
    time_start = time_start + timedelta(seconds=int(time_data[0]/1000))
    time['data'] = time_data/1000. - int(time_data[0]/1000)
    time['units'] = make_time_unit_str(time_start)

    # range
    _range = filemetadata('range')
    max_ngates_scan_index = ngates_per_scan.index(ngates)
    scan_dic = scan_info[max_ngates_scan_index]
    max_ngates_moment_index = scan_dic['ngates'].index(ngates)
    distance_var = scan_dic['distance_vars'][max_ngates_moment_index]
    _range['data'] = dvars[distance_var][:]
    _range['meters_to_center_of_first_gate'] = _range['data'][0]
    _range['meters_between_gates'] = _range['data'][1] - _range['data'][0]

    # fields
    fields = {}
    for moment_name, moment_data in fdata.items():
        field_name = filemetadata.get_field_name(moment_name)
        field_dic = filemetadata(field_name)
        field_dic['_FillValue'] = get_fillvalue()
        field_dic['data'] = moment_data
        fields[field_name] = field_dic

    # metadata
    metadata = filemetadata('metadata')
    metadata['original_container'] = 'NEXRAD Level II'

    # scan_type
    scan_type = 'ppi'

    # latitude, longitude, altitude
    latitude = filemetadata('latitude')
    longitude = filemetadata('longitude')
    altitude = filemetadata('altitude')

    # use the locations in the NetCDF file is available
    if ((hasattr(dataset, 'StationLatitude') and
         hasattr(dataset, 'StationLongitude') and
         hasattr(dataset, 'StationElevationInMeters'))):
        lat = dataset.StationLatitude
        lon = dataset.StationLongitude
        alt = dataset.StationElevationInMeters
    else:
        # if no locations in the file look them up from station name.
        if station is None:
            # determine the station name from the filename
            # this will fail in some cases, in which case station
            # should be implicitly provided in the function call.
            station = os.path.basename(filename)[:4].upper()
        lat, lon, alt = get_nexrad_location(station)
    latitude['data'] = np.array([lat], dtype='float64')
    longitude['data'] = np.array([lon], dtype='float64')
    altitude['data'] = np.array([alt], dtype='float64')

    # sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index
    # sweep_end_ray_index
    sweep_number = filemetadata('sweep_number')
    sweep_mode = filemetadata('sweep_mode')
    sweep_start_ray_index = filemetadata('sweep_start_ray_index')
    sweep_end_ray_index = filemetadata('sweep_end_ray_index')

    sweep_number['data'] = np.arange(nsweeps, dtype='int32')
    sweep_mode['data'] = np.array(
        nsweeps * ['azimuth_surveillance'], dtype='S')
    rays_per_scan = list(radials_per_scan)
    sweep_end_ray_index['data'] = np.cumsum(rays_per_scan, dtype='int32') - 1

    rays_per_scan.insert(0, 0)
    sweep_start_ray_index['data'] = np.cumsum(rays_per_scan[:-1],
                                              dtype='int32')

    # azimuth, elevation, fixed_angle
    azimuth = filemetadata('azimuth')
    elevation = filemetadata('elevation')
    fixed_angle = filemetadata('fixed_angle')
    azimuth['data'] = azim_data
    elevation['data'] = elev_data
    fixed_angle['data'] = fixed_agl_data

    dataset.close()
    return Radar(
        time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=None)


def _scan_info(dvars):
    """ Return a list of information on the scans in the volume. """
    # determine the time of the sweep start
    time_variables = [k for k in dvars.keys() if k.startswith('time')]

    scan_start_times = set([])
    for var in time_variables:
        for time in dvars[var][:, 0]:
            scan_start_times.add(time)
    scan_start_times = list(scan_start_times)
    scan_start_times.sort()

    # build the scan_info list
    time_var_to_moment = {          # time variable to moment conversion
        'timeR': 'Reflectivity',
        'timeV': 'RadialVelocity',
        'timeD': 'DifferentialReflectivity',
        'timeC': 'CorrelationCoefficient',
        'timeP': 'DifferentialPhase',
        'timeR_HI': 'Reflectivity_HI',
        'timeV_HI': 'RadialVelocity_HI',
        'timeD_HI': 'DifferentialReflectivity_HI',
        'timeC_HI': 'CorrelationCoefficient_HI',
        'timeP_HI': 'DifferentialPhase_HI',
    }

    scan_info = [{'start_time': t, 'time_vars': [], 'moments': [],
                  'nradials': [], 'ngates': [], 'elevation_vars': [],
                  'azimuth_vars': [], 'distance_vars': [], 'index': []}
                 for t in scan_start_times]

    for time_var in time_variables:
        for time_var_i, time in enumerate(dvars[time_var][:, 0]):

            scan_index = scan_start_times.index(time)
            scan_dic = scan_info[scan_index]
            moment = time_var_to_moment[time_var]
            _populate_scan_dic(scan_dic, time_var, time_var_i, moment, dvars)

            # corner cases, timeV is a dimension for RadialVelocity AND
            # SpectrumWidth
            if time_var == 'timeV':
                _populate_scan_dic(scan_dic, time_var, time_var_i,
                                   'SpectrumWidth', dvars)
            if time_var == 'timeV_HI':
                _populate_scan_dic(scan_dic, time_var, time_var_i,
                                   'SpectrumWidth_HI', dvars)

    return scan_info


def _populate_scan_dic(scan_dic, time_var, time_var_i, moment, dvars):
    """ Populate a dictionary in the scan_info list. """
    if time_var.endswith('HI'):
        var_suffix = time_var[-4:]
    else:
        var_suffix = time_var[-1:]

    ngates = dvars['numGates' + var_suffix][time_var_i]
    nradials = dvars['numRadials' + var_suffix][time_var_i]

    scan_dic['time_vars'].append(time_var)
    scan_dic['index'].append(time_var_i)
    scan_dic['moments'].append(moment)
    scan_dic['elevation_vars'].append('elevation' + var_suffix)
    scan_dic['azimuth_vars'].append('azimuth' + var_suffix)
    scan_dic['distance_vars'].append('distance' + var_suffix)
    scan_dic['ngates'].append(ngates)
    scan_dic['nradials'].append(nradials)
    return


def _get_moment_data(moment_var, index, ngates):
    """ Retieve moment data for a given scan. """

    # mask, scale and offset
    moment_var.set_auto_maskandscale(False)
    raw_moment_data = moment_var[index][:, :ngates]
    if '_Unsigned' in moment_var.ncattrs():
        if raw_moment_data.dtype == np.int8:
            raw_moment_data = raw_moment_data.view('uint8')
        if raw_moment_data.dtype == np.int16:
            raw_moment_data = raw_moment_data.view('uint16')

    raw_moment_data = np.ma.masked_less_equal(raw_moment_data, 1)

    if 'scale_factor' in moment_var.ncattrs():
        scale = moment_var.scale_factor
    else:
        scale = 1.0

    if 'add_offset' in moment_var.ncattrs():
        add_offset = moment_var.add_offset
    else:
        add_offset = 0.0

    return raw_moment_data * scale + add_offset



""" Unit Tests for Py-ART's graph/cm.py module. """


import matplotlib

from pyart.graph import cm


def test_colormaps_exist():
    assert isinstance(cm.NWSRef, matplotlib.colors.Colormap)
    assert isinstance(cm.NWSRef_r, matplotlib.colors.Colormap)


def test_colormaps_registered():
    cmap = matplotlib.cm.get_cmap('pyart_NWSRef')
    assert isinstance(cmap, matplotlib.colors.Colormap)

    cmap = matplotlib.cm.get_cmap('pyart_NWSRef_r')
    assert isinstance(cmap, matplotlib.colors.Colormap)


def test_tuple_spec():
    # None of the Py-ART colormaps use the tuple spec, so for coverage in the
    # unit tests ddefine a colormap using a tuple

    # data borrowed from matplotlib's lib/matplotlib/_cm.py
    _seismic_data = (
        (0.0, 0.0, 0.3), (0.0, 0.0, 1.0),
        (1.0, 1.0, 1.0), (1.0, 0.0, 0.0),
        (0.5, 0.0, 0.0))
    spec_reversed = cm._reverse_cmap_spec(_seismic_data)

    spec = ((0, 1), (0, 1))
    spec_reversed = cm._reverse_cmap_spec(spec)

    cm.datad['foo'] = _seismic_data
    cmap = cm._generate_cmap('foo', 1)
    assert isinstance(cm.NWSRef_r, matplotlib.colors.Colormap)


def test_revcmap_callable():
    # reversing a callable is not exercised in Py-ART
    data_r = cm.revcmap({'foo': lambda x: x})
    assert data_r['foo'](0) == 1
